unit CRC16;

{16 Bit CRC routines for CCITT polynomial $1021}


interface

(*************************************************************************

 DESCRIPTION     :  16 Bit CRC

 REQUIREMENTS    :  D4-D7/D9-D10/D12, FPC

 EXTERNAL DATA   :  ---

 MEMORY USAGE    :  ---

 DISPLAY MODE    :  ---

 REFERENCES      :  Numerical recipes, 2nd ed.


 Version  Date      Author      Modification
 -------  --------  -------     ------------------------------------------
 0.10     29.09.07  W.Ehrhardt  Initial version based on standard v3.05
 0.11     12.11.08  we          uses BTypes, Ptr2Inc and/or Str255/Str127
 0.12     19.07.09  we          D12 fix: assign with typecast string(fname)
**************************************************************************)

(*-------------------------------------------------------------------------
 (C) Copyright 2002-2009 Wolfgang Ehrhardt

 This software is provided 'as-is', without any express or implied warranty.
 In no event will the authors be held liable for any damages arising from
 the use of this software.

 Permission is granted to anyone to use this software for any purpose,
 including commercial applications, and to alter it and redistribute it
 freely, subject to the following restrictions:

 1. The origin of this software must not be misrepresented; you must not
    claim that you wrote the original software. If you use this software in
    a product, an acknowledgment in the product documentation would be
    appreciated but is not required.

 2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software.

 3. This notice may not be removed or altered from any source distribution.
----------------------------------------------------------------------------*)

{$i STD.INC}

uses
  BTypes;

procedure CRC16Init(var CRC: word);
  {-initialize context}
  {$ifdef DLL} stdcall; {$endif}

procedure CRC16Update(var CRC: word; Msg: pointer; Len: word);
  {-update CRC16 with Msg data}
  {$ifdef DLL} stdcall; {$endif}

procedure CRC16Final(var CRC: word);
  {-CRC16: finalize calculation (dummy)}
  {$ifdef DLL} stdcall; {$endif}

function  CRC16SelfTest: boolean;
  {-Self test for CRC16}
  {$ifdef DLL} stdcall; {$endif}

procedure CRC16Full(var CRC: word; Msg: pointer; Len: word);
  {-CRC16 of Msg with init/update/final}
  {$ifdef DLL} stdcall; {$endif}

procedure CRC16File(const fname: Str255; var CRC: word; var buf; bsize: word; var Err: word);
  {-CRC16 of file, buf: buffer with at least bsize bytes}
  {$ifdef DLL} stdcall; {$endif}


procedure CRC16UpdateXL(var CRC: word; Msg: pointer; Len: longint);
  {-update CRC16 with Msg data}
  {$ifdef DLL} stdcall; {$endif}

procedure CRC16FullXL(var CRC: word; Msg: pointer; Len: longint);
  {-CRC16 of Msg with init/update/final}
  {$ifdef DLL} stdcall; {$endif}


implementation

{CRC table for CCITT polynomial $1021}
const
  ctab: packed array [0..255] of word =
          ( $0000, $1021, $2042, $3063, $4084, $50A5, $60C6, $70E7,
            $8108, $9129, $A14A, $B16B, $C18C, $D1AD, $E1CE, $F1EF,
            $1231, $0210, $3273, $2252, $52B5, $4294, $72F7, $62D6,
            $9339, $8318, $B37B, $A35A, $D3BD, $C39C, $F3FF, $E3DE,
            $2462, $3443, $0420, $1401, $64E6, $74C7, $44A4, $5485,
            $A56A, $B54B, $8528, $9509, $E5EE, $F5CF, $C5AC, $D58D,
            $3653, $2672, $1611, $0630, $76D7, $66F6, $5695, $46B4,
            $B75B, $A77A, $9719, $8738, $F7DF, $E7FE, $D79D, $C7BC,
            $48C4, $58E5, $6886, $78A7, $0840, $1861, $2802, $3823,
            $C9CC, $D9ED, $E98E, $F9AF, $8948, $9969, $A90A, $B92B,
            $5AF5, $4AD4, $7AB7, $6A96, $1A71, $0A50, $3A33, $2A12,
            $DBFD, $CBDC, $FBBF, $EB9E, $9B79, $8B58, $BB3B, $AB1A,
            $6CA6, $7C87, $4CE4, $5CC5, $2C22, $3C03, $0C60, $1C41,
            $EDAE, $FD8F, $CDEC, $DDCD, $AD2A, $BD0B, $8D68, $9D49,
            $7E97, $6EB6, $5ED5, $4EF4, $3E13, $2E32, $1E51, $0E70,
            $FF9F, $EFBE, $DFDD, $CFFC, $BF1B, $AF3A, $9F59, $8F78,
            $9188, $81A9, $B1CA, $A1EB, $D10C, $C12D, $F14E, $E16F,
            $1080, $00A1, $30C2, $20E3, $5004, $4025, $7046, $6067,
            $83B9, $9398, $A3FB, $B3DA, $C33D, $D31C, $E37F, $F35E,
            $02B1, $1290, $22F3, $32D2, $4235, $5214, $6277, $7256,
            $B5EA, $A5CB, $95A8, $8589, $F56E, $E54F, $D52C, $C50D,
            $34E2, $24C3, $14A0, $0481, $7466, $6447, $5424, $4405,
            $A7DB, $B7FA, $8799, $97B8, $E75F, $F77E, $C71D, $D73C,
            $26D3, $36F2, $0691, $16B0, $6657, $7676, $4615, $5634,
            $D94C, $C96D, $F90E, $E92F, $99C8, $89E9, $B98A, $A9AB,
            $5844, $4865, $7806, $6827, $18C0, $08E1, $3882, $28A3,
            $CB7D, $DB5C, $EB3F, $FB1E, $8BF9, $9BD8, $ABBB, $BB9A,
            $4A75, $5A54, $6A37, $7A16, $0AF1, $1AD0, $2AB3, $3A92,
            $FD2E, $ED0F, $DD6C, $CD4D, $BDAA, $AD8B, $9DE8, $8DC9,
            $7C26, $6C07, $5C64, $4C45, $3CA2, $2C83, $1CE0, $0CC1,
            $EF1F, $FF3E, $CF5D, $DF7C, $AF9B, $BFBA, $8FD9, $9FF8,
            $6E17, $7E36, $4E55, $5E74, $2E93, $3EB2, $0ED1, $1EF0  );




{---------------------------------------------------------------------------}
procedure CRC16Update(var CRC: word; Msg: pointer; Len: word);
  {-update CRC16 with Msg data}
begin
  CRC16UpdateXL(CRC, Msg, Len);
end;


{---------------------------------------------------------------------------}
procedure CRC16UpdateXL(var CRC: word; Msg: pointer; Len: longint);
   {-update CRC16 with Msg data}
var
  i: longint;
begin
  for i:=1 to Len do begin
    CRC := CTab[(CRC shr 8) xor PByte(Msg)^] xor ((CRC and $FF) shl 8);
    inc(Ptr2Inc(Msg));
  end;
end;


{---------------------------------------------------------------------------}
procedure CRC16FullXL(var CRC: word; Msg: pointer; Len: longint);
  {-CRC of Msg with init/update/final}
begin
  CRC16Init(CRC);
  CRC16UpdateXL(CRC, Msg, Len);
  CRC16Final(CRC);
end;


{---------------------------------------------------------------------------}
procedure CRC16Init(var CRC: word);
  {-CRC initialization}
begin
  CRC := 0;
end;


{---------------------------------------------------------------------------}
procedure CRC16Final(var CRC: word);
  {-CRC16: finalize calculation (dummy)}
begin
end;


{---------------------------------------------------------------------------}
function  CRC16SelfTest: boolean;
  {-Test from Numerical recipes, 2nd ed. p. 901: CRC('CatMouse987654321')=$E556}
const
  s: string[20] = 'CatMouse987654321';
var
  i,CRC,CRCF: word;
begin
  CRC16Full(CRCF, @s[1], length(s));
  CRC16Init(CRC);
  for i:=1 to length(s) do CRC16Update(CRC, @s[i], 1);
  CRC16Final(CRC);
  CRC16SelfTest := (CRC=$E556) and (CRCF=$E556);
end;


{---------------------------------------------------------------------------}
procedure CRC16Full(var CRC: word; Msg: pointer; Len: word);
  {-CRC16 of Msg with init/update/final}
begin
  CRC16Init(CRC);
  CRC16Update(CRC, Msg, Len);
  CRC16Final(CRC);
end;


{$i-} {Force I-}
{---------------------------------------------------------------------------}
procedure CRC16File(const fname: Str255; var CRC: word; var buf; bsize: word; var Err: word);
  {-CRC16 of file, buf: buffer with at least bsize bytes}
var
  fms: byte;
  L: longint;
  f: file;
begin
  fms := FileMode;
  FileMode := 0;
  system.assign(f,{$ifdef D12Plus} string {$endif} (fname));
  system.reset(f,1);
  Err := IOResult;
  FileMode := fms;
  if Err<>0 then exit;
  CRC16Init(CRC);
  L := bsize;
  while (Err=0) and (L=bsize) do begin
    system.blockread(f,buf,bsize,L);
    Err := IOResult;
    CRC16UpdateXL(CRC, @buf, L);
  end;
  system.close(f);
  if IOResult=0 then;
  CRC16Final(CRC);
end;

end.
